import 'package:geiger_localstorage/geiger_localstorage.dart';

/// <p>A non-persisting dummy mapper for test purposes.</p>
class DummyMapper extends AbstractMapper {
  final Map<String, Node> nodes = {};
  StorageController? controller;
  final String _owner;

  DummyMapper(this._owner);

  @override
  void setController(StorageController controller) {
    this.controller = controller;
  }

  @override
  Future<Node> get(String path) async {
    checkPath(path);
    getSanity(path);
    var ret = nodes[path];
    if (ret == null) {
      throw StorageException('Node "$path" not found');
    }
    if (ret.tombstone) {
      Node returnNode =
          await NodeImpl.fromPath(path, _owner, isTombstone: true);
      returnNode.visibility = ret.visibility;
      return returnNode;
    }
    return ret.deepClone();
  }

  @override
  Future<void> add(Node node) async {
    checkPath(node.path);
    if (nodes[node.path] != null && nodes[node.path]!.tombstone == false) {
      throw StorageException('Node does already exist');
    }
    if (node.skeleton) {
      throw StorageException('Skeleton nodes cannot be added.');
    }
    String? pname = node.parentPath == '' ? ':' : node.parentPath;
    if (node.parentPath != null) {
      if (nodes[pname] == null) {
        throw StorageException(
            ('Parent node "' + (node.parentPath!)) + '" does not exist');
      }
      await nodes[pname]?.addChild(node);
    }
    nodes[node.path] = await node.shallowClone();

    // add children if not skeletonized (not needed... covered by controller)
    //for (Node n in (await node.getChildren()).values) {
    //  if (!n.skeleton) {
    //    add(n);
    //  }
    //}
  }

  @override
  Future<void> update(Node node) async {
    checkPath(node.path);
    if ((!('' == node.parentPath)) && (nodes[node.parentPath] == null)) {
      throw StorageException('Parent node ${node.parentPath} does not exist');
    }
    await nodes[node.path]?.update(node);
  }

  @override
  Future<void> rename(String oldPath, String newPathOrName) async {
    checkPath(oldPath);
    checkPath(newPathOrName);
    Node? oldNode = nodes[oldPath];
    if (oldNode == null) {
      throw StorageException(
          'Old node "$oldPath" does not exist.. therefore, unable to rename to "$newPathOrName"');
    }
    Node newNode = await NodeImpl.fromPath(newPathOrName, _owner);
    newNode.owner = oldNode.owner;
    newNode.visibility = oldNode.visibility;
    for (NodeValue nv in (await oldNode.getValues()).values) {
      await newNode.addValue(nv);
    }
    await add(newNode);
    for (var n in (await oldNode.getChildren()).values) {
      await rename(
          n.path, (newNode.path + GenericController.pathDelimiter) + n.name);
    }
    await delete(oldNode.path);
  }

  @override
  Future<Node> delete(String path) async {
    Node? node = nodes[path];
    if (node == null) {
      throw StorageException('unable to delete node "$path"... does not exist');
    }
    if ((await node.getChildren()).isNotEmpty) {
      throw StorageException(
          'Node does have children... cannot remove ' + path);
    }
    nodes.remove(node);
    Node tombstone =
        await NodeImpl.fromPath(node.path, _owner, isTombstone: true);
    tombstone.visibility = node.visibility;
    nodes[node.path] = tombstone;
    if ((node.parentPath != null) && (!('' == node.parentPath))) {
      await nodes[node.parentPath]!.removeChild(node.name);
    }
    return node;
  }

  @override
  Future<NodeValue?> getValue(String path, String key) async {
    var ret = await get(path);
    if (ret.tombstone) {
      throw StorageException('node does not exist');
    }
    return (await ret.getValues())[key];
  }

  @override
  Future<List<Node>> search(SearchCriteria criteria) async {
    List<Node> l = <Node>[];
    for (var e in nodes.entries) {
      if (await criteria.evaluate(e.value)) {
        l.add(e.value);
      }
    }
    return l;
  }

  @override
  Future<void> close() async {}

  @override
  Future<void> flush() async {}

  @override
  Future<void> zap() async {
    nodes.clear();
  }
}
