library geiger_localstorage;

import 'dart:convert';

import '../../geiger_localstorage.dart';

/// <p>Helper class for serialization serializes important java primitives.</p>
class SerializerHelper {
  static const int stringUid = 123798371293;
  static const int longUid = 1221312393;
  static const int intUid = 122134568793;
  static const int stacktraceUid = 9012350123956;

  static const int intSize = 4;
  static const int longSize = 8;
  static const int _byteSize = 8;

  static void _writeInteger(ByteSink out, int l, int size) {
    List<int> result = intToByteArray(l, size);
    out.sink.add(result);
  }

  static Future<int> _readInteger(ByteStream in_, int size) async {
    List<int> arr = await in_.popArray(size);
    return byteArrayToInt(arr, size);
  }

  static void writeRawInt(ByteSink out, int l) async {
    _writeInteger(out, l, intSize);
  }

  static Future<int> readRawInt(ByteStream in_) async {
    return _readInteger(in_, intSize);
  }

  static void writeRawLong(ByteSink out, int l) async {
    _writeInteger(out, l, longSize);
  }

  static Future<int> readRawLong(ByteStream in_) async {
    return _readInteger(in_, longSize);
  }

  static Future<List<int>> peekRawLong(ByteStream in_, int num) async {
    ByteStream bs = ByteStream(null, await in_.peekBytes(longSize * num));
    List<int> ret = [];
    for (int i = 0; i < num; i++) {
      ret.add(await _readInteger(bs, longSize));
    }
    return ret;
  }

  /// Convert the first [size] bytes of List<int> [bytes] to a signed int.
  static int byteArrayToInt(List<int> bytes, [int size = intSize]) {
    int result = 0;
    for (int i = 0; i < size; i++) {
      result |= (bytes[i] << (_byteSize * i));
    }
    // calculate two complement if necessary
    if (result >> (size * _byteSize - 1) & 1 == 1) {
      // this is a negative number
      result = result ^ ((1 << (size * _byteSize)) - 1);
      result *= -1;
      result -= 1;
    }
    return result;
  }

  /// Convert int [value] to bytearray of length [size].
  static List<int> intToByteArray(int value, [int size = intSize]) {
    List<int> result = [];
    for (int i = 0; i < size; i++) {
      int i1 = (value & ((1 << _byteSize) - 1));
      result.add(i1);
      value >>= _byteSize;
    }
    return result;
  }

  /// <p>Serialize a long variable.</p>
  /// @param out the stream to be read
  /// @param l   the value to be deserialized
  /// @throws IOException if an exception occurs while writing to the stream
  static void writeLong(ByteSink out, int l) {
    writeRawLong(out, longUid);
    writeRawLong(out, l);
  }

  /// <p>Deserialize a long variable.</p>
  /// @param in the stream to be read
  /// @return the deserialized long value
  /// @throws IOException if an exception occurs while writing to the stream
  static Future<int> readLong(ByteStream in_) async {
    castTest('long', longUid, await readRawLong(in_));
    return readRawLong(in_);
  }

  static Future<int> peekLong(ByteStream in_) async {
    List<int> longs = await peekRawLong(in_, 2);
    castTest('long', longUid, longs[1]);
    return longs[2];
  }

  /// <p>Serialize an int variable.</p>
  /// @param out the stream to be read
  /// @param i   the value to be deserialized
  /// @throws IOException if an exception occurs while writing to the stream
  static void writeInt(ByteSink out, int i) {
    writeRawLong(out, intUid);
    writeRawInt(out, i);
  }

  /// <p>Deserialize an int variable.</p>
  /// @param in the stream to be read
  /// @return the deserialized integer value
  /// @throws IOException if an exception occurs while writing to the stream
  static Future<int> readInt(ByteStream in_) async {
    castTest('int', intUid, await readRawLong(in_));
    return readRawInt(in_);
  }

  /// <p>Serialize a string variable.</p>
  /// @param out the stream to be read
  /// @param s   the value to be deserialized
  /// @throws IOException if an exception occurs while writing to the stream
  static void writeString(ByteSink out, String? s) {
    SerializerHelper.writeRawLong(out, stringUid);
    if (s == null) {
      writeRawInt(out, -1);
    } else {
      List<int> barr = utf8.encode(s);
      writeRawInt(out, barr.length);
      out.sink.add(barr);
    }
  }

  /// <p>Deserialize a string variable.</p>
  /// @param in the stream to be read
  /// @return the deserialized string
  /// @throws IOException if an exception occurs while writing to the stream
  static Future<String?> readString(ByteStream in_) async {
    castTest('String', stringUid, await readRawLong(in_));
    int length = await readRawInt(in_);
    if (length == (-1)) {
      return null;
    } else {
      List<int> arr = await in_.popArray(length);
      return utf8.decode(arr);
    }
  }

  /// <p>Serialize an array of StackTraces.</p>
  /// @param out the stream to be read
  /// @param ste the value to be deserialized
  /// @throws IOException if an exception occurs while writing to the stream
  static void writeStackTraces(ByteSink out, StackTrace? ste) {
    writeRawLong(out, stacktraceUid);
    if (ste == null) {
      writeRawInt(out, -1);
    } else {
      writeString(out, ste.toString());
    }
  }

  /// <p>Deserialize an array of StackTraceElement variable.</p>
  /// @param in the stream to be read
  /// @return the deserialized array
  /// @throws IOException if an exception occurs while writing to the stream
  static Future<StackTrace?> readStackTraces(ByteStream in_) async {
    castTest('StackTrace', stacktraceUid, await readRawLong(in_));
    int length = await readRawInt(in_);
    if (length == (-1)) {
      return null;
    } else {
      return StackTrace.fromString(await readString(in_) ?? '');
    }
  }

  /// short casting test.
  ///
  /// Verifies if the [expected] result matches the result we [really] got and
  /// throws an exception if they do not matcch.
  static void castTest(String obj, int expected, int really, [int? num]) {
    if (expected != really) {
      String s =
          'cannot cast to ' + obj + ' (expected: $expected; got: $really)';
      if (num != null) {
        s += ' ($num)';
      }
      throw StorageException(s);
    }
  }

  /// Read an object from ByteArrayInputStream.
  /// @param in the byteArrayInputStream to use
  /// @return return the object read
  /// @throws IOException if object cannot be read
/*    static Object readObject(Stream<List<int>> in_)
    {
        switch (("" + SerializerHelper.readIntLong(in_))) {
            case ("" + SerializerHelper.STRING_UID):
                List<int?> arr = [];
                arr = readIntInt(in_);
                in_.read(arr);
                return utf8.encode(arr);
            case ("" + SerializerHelper.LONG_UID):
                return readIntLong(in_);
            default:
                throw new Exception("Cannot cast");
        }
    }*/

  /// Write an object to ByteArrayOutputStream.
  /// @param out the ByteArrayOutputStream to use
  /// @param o the Object to write
  /// @throws IOException if object cannot be written
/*    static void writeObject(Sink<List<int>> out, dynamic o)
    {
        switch (o.runtimeType.toString()) {
            case "String":
                writeString(out, (String().(o)));
                break;
            case "long":
                writeLong(out, o);
                break;
            case "int":
                writeInt(out, o);
                break;
            default:
                throw new Exception("Cannot cast ${o.runtimeType.toString()}");
        }
    }*/

}
