import 'dart:async';

import 'package:geiger_localstorage/geiger_localstorage.dart';
import 'package:test/test.dart';

class ChangeEvent {
  final EventType type;
  final Node? oldNode;
  final Node? newNode;

  ChangeEvent(this.type, this.oldNode, this.newNode);

  @override
  String toString() {
    return 'ChangeEvent{type: $type, oldNode: $oldNode, newNode: $newNode}';
  }
}

class _ConditionalCompleter {
  final bool Function(List<ChangeEvent>) canComplete;
  final List<ChangeEvent> Function(List<ChangeEvent>) createResult;

  final _completer = Completer<List<ChangeEvent>>();

  Future<List<ChangeEvent>> get future => _completer.future;

  _ConditionalCompleter(this.canComplete, this.createResult);

  void complete(List<ChangeEvent> events) {
    if (!canComplete(events)) return;
    _completer.complete(createResult(events));
  }
}

class CollectingListener with StorageListener {
  final List<ChangeEvent> events = [];

  final Set<_ConditionalCompleter> completers = {};

  @override
  Future<void> gotStorageChange(
      EventType event, Node? oldNode, Node? newNode) async {
    events.add(ChangeEvent(event, oldNode, newNode));
    for (final completer in completers) {
      completer.complete(events);
    }
  }

  Future<List<ChangeEvent>> awaitCount(int count,
      [Duration timeLimit = const Duration(seconds: 1)]) {
    canComplete(List events) => events.length >= count;
    createResult(List<ChangeEvent> events) =>
        List<ChangeEvent>.from(events.getRange(0, count));
    if (canComplete(events)) {
      return Future.value(createResult(events));
    }

    final completer = _ConditionalCompleter(canComplete, createResult);
    completers.add(completer);
    return completer.future.timeout(timeLimit, onTimeout: () {
      completers.remove(completer);
      throw TimeoutException(
          'Did not receive enough change events before timeout.', timeLimit);
    });
  }
}

const owner = 'testOwner';

Future<StorageController> getNewController() async {
  var controller = GenericController(owner, DummyMapper(owner));
  await controller.zap();
  await ExtendedTimestamp.initializeTimestamp(controller);
  return controller;
}

void main() {
  group('change listener', () {
    test('register/deregister', () async {
      final controller = await getNewController();
      final listener = CollectingListener();
      final criteria = SearchCriteria();

      await controller.registerChangeListener(listener, criteria);
      // test overwrite
      await controller.registerChangeListener(listener, criteria);
      var result = await controller.deregisterChangeListener(listener);

      expect(result.length, 1);
      expect(result.first, criteria);

      // test deregister unregistered listener
      expect(() => controller.deregisterChangeListener(listener),
          throwsA(const TypeMatcher<StorageException>()));
    });

    group('event types', () {
      testEventType(
          EventType expectedType,
          Node? expectedOldNode,
          Future<void> Function(StorageController) setup,
          Future<Node?> Function(StorageController) execute) {
        test(expectedType.toValueString(), () async {
          final controller = await getNewController();
          final listener = CollectingListener();
          final criteria = SearchCriteria();

          await setup(controller);
          await controller.registerChangeListener(listener, criteria);
          final expectedNewNode = await execute(controller);

          await listener.awaitCount(1);
          expect(listener.events.length, 1);
          final event = listener.events.first;
          expect(event.type, expectedType);
          expect(event.oldNode, expectedOldNode);
          expect(event.newNode, expectedNewNode);
        });
      }

      {
        final node = NodeImpl(':testNode', owner);
        testEventType(EventType.create, null, (_) async {}, (controller) async {
          await controller.add(node);
          return node;
        });
      }
      {
        const path = ':testNode';
        final oldNode = NodeImpl(path, owner);
        final newNode = NodeImpl(path, owner, null, Visibility.amber);
        testEventType(EventType.update, oldNode, (controller) async {
          await controller.add(oldNode);
        }, (controller) async {
          await controller.update(newNode);
          return newNode;
        });
      }
      {
        const path = ':testNode';
        final node = NodeImpl(path, owner);
        testEventType(EventType.delete, node, (controller) async {
          await controller.add(node);
        }, (controller) async {
          await controller.delete(path);
          return null;
        });
      }
      {
        const oldPath = ':oldTestNode';
        const newPath = ':newTestNode';
        final node = NodeImpl(oldPath, owner);
        testEventType(EventType.rename, node, (controller) async {
          await controller.add(node);
        }, (controller) async {
          await controller.rename(oldPath, newPath);
          return await controller.get(newPath);
        });
      }
    });

    test('with criteria', () async {
      const path = ':testNode';
      final node = NodeImpl(path, owner);
      final otherNode = NodeImpl(':otherTestNode', owner);

      final controller = await getNewController();
      final listener = CollectingListener();
      final criteria = SearchCriteria(searchPath: path);

      await controller.registerChangeListener(listener, criteria);
      await controller.add(otherNode);
      await controller.add(node);
      otherNode.lastModified = 1640991600000;
      await controller.update(otherNode);

      await listener.awaitCount(1);
      expect(listener.events.length, 1);
      expect(listener.events.first.newNode?.path, path,
          reason: 'Received event for the wrong node.');
    });
  });
}
